#!/usr/bin/env bash
TEST_DIR=$(cd $(dirname $0) && pwd)
HOSTLIST=$TEST_DIR/${HOSTFILE:-"hostfile"}

if [ -n "$WORLD_SIZE" ]; then
  NNODES=$WORLD_SIZE
elif [ -f $HOSTLIST ]; then
  NNODES=$(cat $HOSTLIST | wc -l)
else
  NNODES=${NNODES:-"1"}
fi

if [ -n "$RANK" ]; then
  NODE_RANK=$RANK
elif [ -f $HOSTLIST ]; then
  ME=$(hostname -I | awk '{print $1}')
  NODE_RANK=$(awk -v me=$ME '$0 == me { print NR-1 }' $HOSTLIST)
else
  NODE_RANK=${NODE_RANK:-"0"}
fi
[ $NODE_RANK -lt $NNODES ] || exit 0

if [ -n "$NPROC_PER_NODE" ]; then
  NGPUS_PER_NODE=$NPROC_PER_NODE
elif [ -x "$(command -v nvidia-smi)" ]; then
  NGPUS_PER_NODE=$(nvidia-smi --query-gpu=count --format=csv,noheader | head -n1)
else
  NGPUS_PER_NODE=1
fi

if [ -z "$MASTER_ADDR" ]; then
  if [ -f $HOSTLIST ]; then
    MASTER_ADDR=$(head -n1 $HOSTLIST)
  else
    MASTER_ADDR=127.0.0.1
  fi
fi

MASTER_PORT=${MASTER_PORT:-"49152"}

RESULTS_DIR=$TEST_DIR/results
CHECKPOINTS_DIR=$TEST_DIR/checkpoints
[ -d $RESULTS_DIR ] || mkdir -p $RESULTS_DIR
[ -d $CHECKPOINTS_DIR ] || mkdir -p $CHECKPOINTS_DIR

PHASE1="\
  --train_batch_size=${BATCH_SIZE:-128} \
  --learning_rate=${LR:-6e-3} \
  --warmup_proportion=${WARMUP_UPDATES:-0.2843} \
  --input_dir=$TEST_DIR/data/hdf5_lower_case_1_seq_len_128_max_pred_20_masked_lm_prob_0.15_random_seed_12345_dupe_factor_5/wikicorpus_en \
  --max_seq_length=128 --max_predictions_per_seq=20 \
  --max_steps=${MAX_STEPS:-7038} \
  --skip_checkpoint \
  --num_steps_per_checkpoint=2500 \
  "

PHASE2="\
  --train_batch_size=${BATCH_SIZE:-24} \
  --learning_rate=${LR:-4e-3} \
  --warmup_proportion=${WARMUP_UPDATES:-0.128} \
  --input_dir=$TEST_DIR/data/hdf5_lower_case_1_seq_len_512_max_pred_80_masked_lm_prob_0.15_random_seed_12345_dupe_factor_5/wikicorpus_en \
  --phase2 --resume_from_checkpoint --phase1_end_step=7038 \
  --max_seq_length=512 --max_predictions_per_seq=80 \
  --max_steps=${MAX_STEPS:-1563} \
  --skip_checkpoint \
  --num_steps_per_checkpoint=1000 \
  "

PHASES=( "$PHASE1" "$PHASE2" )
PHASE=${PHASE:-"1"}

shopt -s extglob
export LD_LIBRARY_PATH=${LD_LIBRARY_PATH//\/usr\/local\/nvidia\/lib?([6])?([4]):}

if [ -n "$CONDA_ENV" ]; then
  source /opt/conda/bin/activate $CONDA_ENV
fi
ACCL_HOST_ID=$NODE_RANK \
python -m torch.distributed.launch \
  --nnodes=$NNODES --node_rank=$NODE_RANK --nproc_per_node=$NGPUS_PER_NODE \
  --master_addr=$MASTER_ADDR --master_port=$MASTER_PORT \
  $TEST_DIR/run_pretraining.py \
    --backend=${BACKEND:-"accl"} \
    --ddp_type=${DDP_TYPE:-"native"} --bucket_size=${BUCKET_SIZE:-"25"} \
    --bert_model=bert-large-uncased \
    --do_train ${PHASES[$((PHASE-1))]} --fp16 --seed=12439 \
    --gradient_accumulation_steps=1 \
    --config_file=$TEST_DIR/bert_config.json --output_dir=$CHECKPOINTS_DIR \
    --json-summary $RESULTS_DIR/dllogger.json
